import autoarray.plot as aplt


class HalfLightRadiusAXVLine(aplt.AXVLine):
    pass


class EinsteinRadiusAXVLine(aplt.AXVLine):
    pass


class ModelFluxesYXScatter(aplt.YXScatter):
    pass


class LightProfileCentresScatter(aplt.GridScatter):
    pass


class MassProfileCentresScatter(aplt.GridScatter):
    pass


class MultipleImagesScatter(aplt.GridScatter):
    pass


class CriticalCurvesPlot(aplt.GridPlot):
    pass


class CausticsPlot(aplt.GridPlot):
    pass
